#include "shader.h"
#include <glew/glew.h>
#include <iostream>
#include "glassert.h"
#include <glm/ext.hpp>

const std::string Shader::DEFAULT_VERT = "#version 330\nlayout(location = 0)in vec3 inPosition;\nvoid main(){gl_Position = vec4(inPosition, 1.0);}";
const std::string Shader::DEFAULT_FRAG = "#version 330\nlayout(location = 0)out vec4 outColor;\nvoid main(){outColor = vec4(1, 0, 0, 1);}";

Shader::Shader(const QString* vertexSource, const QString* fragmentSource)
{
    program = glAssert(glCreateProgram());

    GLuint vertexShaderId = createShader(vertexSource, GL_VERTEX_SHADER);
    GLuint fragmentShaderId = createShader(fragmentSource, GL_FRAGMENT_SHADER);

    glAssert(glAttachShader(program, vertexShaderId));
    glAssert(glAttachShader(program, fragmentShaderId));

    glAssert(glBindAttribLocation(program, 0, "inPosition"));
    glAssert(glBindAttribLocation(program, 1, "inNormal"));
    glAssert(glBindAttribLocation(program, 2, "inTexCoord"));

    glAssert(glLinkProgram(program));

    // check errors
    GLint linked;
    glAssert(glGetProgramiv(program, GL_LINK_STATUS, &linked));
    if (!linked) {
        std::cerr << "Program not linked" << std::endl;
        printProgramInfoLog(program);
        program = 0;
    }
    else
        std::cout << "Shader successfully compiled" << std::endl;

    glAssert(glDetachShader(program, vertexShaderId));
    glAssert(glDetachShader(program, fragmentShaderId));
    glAssert(glDeleteShader(vertexShaderId));
    glAssert(glDeleteShader(fragmentShaderId));
}

Shader::~Shader()
{
    bool ok;
    glAssert(ok = glIsProgram(program))
    if(ok)
        glAssert(glDeleteProgram(program));
}

GLuint Shader::createShader(const QString* source, GLenum shaderType)
{
    glAssert(GLuint shaderId = glCreateShader(shaderType));
    const GLchar *data = (const GLchar *)source->toStdString().c_str();
    glAssert(glShaderSource(shaderId, 1, &data, NULL));

    glAssert(glCompileShader(shaderId));

    // check errors
    GLint compiled;
    glAssert(glGetShaderiv(shaderId, GL_COMPILE_STATUS, &compiled));
    if (!compiled) {
        std::cerr << (shaderType == GL_VERTEX_SHADER ? "Vertex" : "Fragment") << "shader not compiled : " << std::endl;
        printShaderInfoLog(shaderId);
        return 0;
    }
    else
        return shaderId;
}

void Shader::printShaderInfoLog(GLuint shaderId)
{
    int infoLogLen = 0;
    int charsWritten = 0;
    GLchar *infoLog;
    glAssert(glGetShaderiv(shaderId, GL_INFO_LOG_LENGTH, &infoLogLen));
    if (infoLogLen > 0) {
        infoLog = new GLchar[infoLogLen];
        glAssert(glGetShaderInfoLog(shaderId, infoLogLen, &charsWritten, infoLog));
        std::cerr << "InfoLog:" << std::endl << infoLog << std::endl;
        delete [] infoLog;
    }
}

void Shader::printProgramInfoLog(GLuint programId)
{
    int infoLogLen = 0;
    int charsWritten = 0;
    GLchar *infoLog;
    glAssert(glGetProgramiv(programId, GL_INFO_LOG_LENGTH, &infoLogLen));
    if (infoLogLen > 0) {
        infoLog = new GLchar[infoLogLen];
        glAssert(glGetProgramInfoLog(programId, infoLogLen, &charsWritten, infoLog));
        std::cerr << "InfoLog:" << std::endl << infoLog << std::endl;
        delete [] infoLog;
    }
}

GLuint Shader::getLocation(std::string attribName)
{
    glAssert(GLuint loc = glGetUniformLocation(program, attribName.c_str()));
    return loc;
}

void Shader::bind()
{
    glAssert(glUseProgram(program));
}

void Shader::unbind()
{
    glAssert(glUseProgram(0));
}

void Shader::bindFloat(GLuint location, float val)
{
    glAssert(glUniform1f(location, val));
}

void Shader::bindMatrix(GLuint location, glm::mat4 mat)
{
    glAssert(glUniformMatrix4fv(location, 1, GL_FALSE, glm::value_ptr(mat)));
}

void Shader::bindVec3(GLuint location, glm::vec3 vec)
{
    glAssert(glUniform3fv(location, 1, glm::value_ptr(vec)));
}

void Shader::bindTexture(GLuint location, GLuint tex_id)
{
    glAssert(glUniform1i(location, tex_id));
}
